#! /usr/bin/env python3
import rospy
import os
import hanqiangpose
import command
import subprocess
import hjsx
import threading
import numpy as np
import multiprocessing
from moveit_msgs.msg import PlanningScene
from dynamic_reconfigure.client import Client
from rospy.exceptions import ROSException
from std_srvs.srv import Empty
from rospy.service import ServiceException

waited_once = False

def wait_for_topic(topic_name, message_type):
    try:
        rospy.loginfo("等待场景地图加载完毕...")
        message = rospy.wait_for_message(topic_name, message_type, timeout=None)
        rospy.loginfo(f"场景地图已加载完毕")
        return message
    except rospy.ROSException as e:
        rospy.logerr(f"等待话题 {topic_name} 时程序被中断: {e}")
        return None
    
def clear_octomap():
    try:
        clear_octomap_service = rospy.ServiceProxy('/clear_octomap', Empty)  # 创建服务代理
        clear_octomap_service()  # 尝试调用服务
        rospy.loginfo("Octomap has been cleared.")
    except ServiceException as e:
        rospy.loginfo("Failed to call /clear_octomap service or it is not available. Skipping operation.")

if __name__ == "__main__":
    rospy.init_node("set_update_paramter_p")

    # 初始化各种类型参数
    rospy.set_param("sequence_ready",0)
    rospy.set_param("sign_control",0)
    rospy.set_param("sign_pointcloud",0)
    rospy.set_param("sign_traj_accepted",0)
    rospy.set_param("yaw",np.pi/6)  #内收角（偏航角）
    rospy.set_param("yaw_rate",100) #偏航角过渡速度（角度变化45度/100mm）
    rospy.set_param("pitch_of_Horizontalweld",np.pi/4)  #和底面夹角（俯仰角）
    rospy.set_param("pitch_of_Verticalweld",np.pi/5)
    rospy.set_param("planner_id","RRTConnect")
    rospy.set_param("pointcloud_topic","/pointcloud/output")
    rospy.set_param("culling_radius",7) #焊缝剔除半径
    # rospy.set_param("/move_group/octomap_resolution",0.006)
    rospy.set_param("folder_path","/home/chen/catkin_ws/src/publish_pointcloud/publish_pointcloud/data")

    # 显示参数当前值
    sign_pointcloud = rospy.get_param("sign_pointcloud")

    rospy.loginfo("当前参数值:")
    rospy.loginfo("sign_pointcloud = %s",sign_pointcloud)
    rospy.loginfo("yaw = %s",rospy.get_param("yaw")/np.pi*180)
    rospy.loginfo("pitch_of_Horizontalweld = %s",rospy.get_param("pitch_of_Horizontalweld")/np.pi*180)
    rospy.loginfo("pitch_of_Verticalweld = %s",rospy.get_param("pitch_of_Verticalweld")/np.pi*180)
    while True:
        sign_control = str(rospy.get_param("sign_control"))
        if sign_control == "0":
            if not waited_once:
                print("等待点云数据准备完成...")
                waited_once = True
        elif sign_control == "1":
            # 初始化参数
            rospy.set_param("sign_control",0)
            rospy.set_param("sign_traj_accepted",0)
            
            # 获取新的文件夹值
            new_folder_path = rospy.get_param("folder_path")
            rospy.loginfo("folder_path:%s",new_folder_path)
            
            # 清除场景
            clear_octomap()

            rospy.loginfo("launch pointcloud node now")
            process = multiprocessing.Process(target=command.launch_publish_pointcloud_background)
            # 启动子进程
            process.start()
            # command.launch_publish_pointcloud()

            hjsx.run()
            hanqiangpose.run()

            # 等待 /move_group/monitored_planning_scene 话题发布消息
            planning_scene = wait_for_topic('/move_group/monitored_planning_scene', PlanningScene)
            rospy.set_param("sign_pointcloud",1)
            rospy.loginfo("终止点云发布，关闭发布窗口")
            
            rospy.loginfo("launch moveitserver node now")
            process = multiprocessing.Process(target=command.launch_moveit_control_server_background)
            process.start()
            
            #运行结束，重置参数
            waited_once = False
 
        elif sign_control == "2":
            # 重置参数
            rospy.set_param("sign_control",0)

            rospy.set_param("planner_id","BiEST")
            rospy.loginfo("launch moveitserver node now")
            process = multiprocessing.Process(target=command.launch_moveit_control_server_background)
            process.start()
            rospy.set_param("planner_id","RRTConnect")
            #运行结束，重置参数
            waited_once = False
        

    